// Spline-1-convolution.cc
//
// Implementation for methods of Spline<1>
//

#include "splines/Spline.h"
#include "splines/Grid.h"
#include "splines/DeltaFunction.h"
#include "splines/GridSearchAnswer.h"
#include <vector>
#include <algorithm>
#include <cassert>
#include <sstream>
#include <iomanip>
using std::setw;

//#define TRACE_C1
//#define TRACE_C2
//#define TRACE_C3
//#define TRACE_C4
//#define TRACE_C5

namespace splines {

static void convolutionAccumulate
        ( double & a, double b, 
          double q,   Spline<1> const & f, Function_1 const & g,
          double & s_a, double & integral )
{
  double midpoint = 0.5*(a+b);
  double s_b = f(q-b)*g(b);
  #ifdef TRACE_C3
  std::cout << "integral coming into accumlate was: " << integral << "\n";
  #endif
  integral += ((b-a)/6.0) * (s_a+4*f(q-midpoint)*g(midpoint)+s_b); 
  #ifdef TRACE_C3
  std::cout << " (a,b) = (" << a << "," << b << ") "
            << "f(q-b) = " << f(q-b) << " g(b) = " << g(b)  << "\n"
            << "          s_a = " << s_a << " s_b = " << s_b 
            << " integral = " << integral << "\n";
  #endif
  a = b;
  s_a = s_b;
}          

Spline<1>
Spline<1>::convolution(Spline<1> const & g) const
{
  // $$$ See Mathematical Details TODO.TODO $$$
  
  // Return a Spline<1> using the same grid as this, smearing by 
  // doing s(q) = integral [f(q-p)g(p)] dp where f(x) is this spline
  //
  // It is assumed that g(p) integrates to 1; if not, normalization is 
  // performed so as to use a smearing function proportional to g which 
  // does integrate to 1.
  
  // TODO -- does not handle embedded spliness for g or f.
  //         These may or may not make sense; the range issue is subtle

  if (g.isDeltaFunction) {
    return deltaFunctionConvolution (g);
  }

  double normalization = 1.0;
  // TODO -- normalization should be 1/integral(g)
  
  std::vector<double> sValues (grid.nPoints());
  
  for (unsigned int qPoint = 0; qPoint < grid.nPoints(); qPoint++) {
//  for (unsigned int qPoint = 0; qPoint < 2; qPoint++) {
    double integral = 0.0;
    double q = grid[qPoint];
    double a = g.grid[0]; // because g is assumed to be 0 at points p < g.grid[0]
    double b = 0.0;       // starting value of b is moot 
    double s_a = f(q-a)*g(a);
    #ifdef TRACE_C2
    std:: cout << "q = " << q << " a = " << a 
               << " f(q-a) = " << f(q-a)
               << " g(a) = " << g(a) << "\n";
    #endif
    // Loop invariants:  
    //   fGridIndex refers to the lowest f point already used in the accumulation
    //              (the lowest since we are using f(q-b) not f(b))
    //   gGridIndex refers to the highest g point already used 
    //   a is the low point in the next interval to do (the old b during loop)
    //   s_a is the value of f(q-a)g(a) 
    // Loop structure -- isomorphic to that of the merge algorithm: (thanks web)
    //   while (we can consume from either list) 
    //      consume the lower of the two values 
    //   while (list g is not empty) consume from list g
    //   while (list f is not empty) consume from list f @
    // @ We are in a special case - if g is empty, even if f is not, the 
    //   accumulation is a null operation, since outside the g range, g=0 
    int fGridIndex = grid.nPoints()-1;
    while ( q-grid[fGridIndex] <= a ) {
      --fGridIndex;
    }
    #ifdef TRACE_C2
    std:: cout << "starting fGridIndex = " << fGridIndex
               << " grid[fGridIndex] = " << grid[fGridIndex]
               << " on g grid " << q-grid[fGridIndex] << "\n";
    #endif
    
    int gGridIndex = 0;
    while ( (fGridIndex >= 0) && (gGridIndex < g.grid.nPoints() ) ) {
      if  ( g.grid[gGridIndex] < q - grid[fGridIndex] ) {
        // consume from g
        b = g.grid[gGridIndex++];
        #ifdef TRACE_C3
        std::cout << "convolutionAccumulate from " 
                  << a << " to " << b << " (in g grid) \n";
        #endif
        convolutionAccumulate ( a, b, q, *this, g, s_a, integral );
        #ifdef TRACE_C3
        std::cout << "integral comes back as "  << integral << "\n";
        #endif
      } else {
        // consume from f
        b = q - grid[fGridIndex--];
        #ifdef TRACE_C3
        std::cout << "convolutionAccumulate from " 
                  << a << " to " << b << " (in f grid) \n";
        #endif 
        convolutionAccumulate ( a, b, q, *this, g, s_a, integral );
        #ifdef TRACE_C3
        std::cout << "integral comes back as "  << integral << "\n";
        #endif
      }
    }
    #ifdef TRACE_C3
    std::cout << "finishing g grid \n";
    #endif 
    while  ( gGridIndex < g.grid.nPoints() ) {
      // consume from g
      b = g.grid[gGridIndex++];
      #ifdef TRACE_C3
      std::cout << "convolutionAccumulate from " 
                << a << " to " << b << " (in g grid) \n";
      #endif 
      convolutionAccumulate ( a, b, q, *this, g, s_a, integral );
    }
    sValues[qPoint] = integral * normalization;           
    #ifdef TRACE_C1
    std::cout << "s[" << qPoint << "] = " << sValues[qPoint] << "\n";
    #endif 
  }
  return Spline<1>( sValues, grid );
  
} // convolution

Spline<1>
Spline<1>::convolution(Function_1 const & g, double g_min, double g_max) const
{
  // $$$ See Mathematical Details TODO.TODO $$$
  
  // Return a Spline<1> using the same grid as this, smearing by 
  // doing s(q) = integral [f(q-p)g(p)] dp where f(x) is this spline
  //
  // It is assumed that g(p) integrates to 1; if not, normalization is 
  // performed so as to use a smearing function proportional to g which 
  // does integrate to 1.
  
  double normalization = 1.0;
  // TODO -- normalization should be 1/integral(g)
  
  std::vector<double> sValues (grid.nPoints());
  
  for (unsigned int qPoint = 0; qPoint < grid.nPoints(); qPoint++) {
//  for (unsigned int qPoint = 0; qPoint < 2; qPoint++) {
    double integral = 0.0;
    double q = grid[qPoint];
    double a = g_min; // because g is assumed to be 0 at points p < g_min
    double b = 0.0;       // starting value of b is moot 
    double s_a = f(q-a)*g(a);
    double endSpacing = 0; // starting value of endSpacing is moot 
    #ifdef TRACE_C2F
    std:: cout << "q = " << q << " a = " << a 
               << " f(q-a) = " << f(q-a)
               << " g(a) = " << g(a) << "\n";
    #endif
    // Loop invariants:  
    //   fGridIndex refers to the lowest f point already used in the accumulation
    //              (the lowest since we are using f(q-b) not f(b))
    //   a is the low point in the next interval to do (the old b during loop)
    //   s_a is the value of f(q-a)g(a) 
    // Loop structure
    //   while (we can consume from f and we are not past g_max) 
    //      consume with b as the lower of the two values 
    //   while (we are not past g_max) 
    //      consume using step size saved from the last f step
    //         
    int fGridIndex = grid.nPoints()-1;
    while ( q-grid[fGridIndex] <= a ) {
      --fGridIndex;
    }
    #ifdef TRACE_C2F
    std:: cout << "starting fGridIndex = " << fGridIndex
               << " grid[fGridIndex] = " << grid[fGridIndex]
               << " on g grid " << q-grid[fGridIndex] << "\n";
    #endif
    
    double gPoint = g_min;
    while ( (fGridIndex >= 0) && (gPoint < g_max ) ) {
      if  ( g_max < q - grid[fGridIndex] ) {
        // "consume" from g
        b = g_max;
        gPoint = g_max;
        #ifdef TRACE_C3F
        std::cout << "convolutionAccumulate from " 
                  << a << " to " << b << " (from g_max) \n";
        #endif
        convolutionAccumulate ( a, b, q, *this, g, s_a, integral );
        #ifdef TRACE_C3F
        std::cout << "integral comes back as "  << integral << "\n";
        #endif
      } else {
        // consume from f
        b = q - grid[fGridIndex--];
        gPoint = b;  // perhaps not strictly necessary, but ...
        endSpacing = b - a;
        #ifdef TRACE_C3F
        std::cout << "convolutionAccumulate from " 
                  << a << " to " << b << " (in f grid) \n";
        #endif 
        convolutionAccumulate ( a, b, q, *this, g, s_a, integral );
        #ifdef TRACE_C3F
        std::cout << "integral comes back as "  << integral << "\n";
        #endif
      }
    }
    #ifdef TRACE_C3F
    std::cout << "finishing g range \n";
    #endif 
    while  ( gPoint < g_max ) {
      // "consume" from g
      b = a + endSpacing; 
      gPoint = std::max(b,g_max);
      #ifdef TRACE_C3
      std::cout << "convolutionAccumulate from " 
                << a << " to " << b << " (in g grid) \n";
      #endif 
      convolutionAccumulate ( a, b, q, *this, g, s_a, integral );
    }
    sValues[qPoint] = integral * normalization;           
    #ifdef TRACE_C1
    std::cout << "s[" << qPoint << "] = " << sValues[qPoint] << "\n";
    #endif 
  }
  return Spline<1>( sValues, grid );
  
} // convolution

Spline<1>
Spline<1>::deltaFunctionConvolution (Spline<1> const & g) const {
  // By intent, we ignore the delta function height, and treat
  // the normalization as unity.  Then
  // f (*) delta(x-x_0) [q] is integral { f(q-p)delata(p-x_0) dp = f(q-x_0)
  
  Spline<1> s(*this);
  s.grid.shift(-g.deltaFunctionLocation);

} // deltaFunctionConvolution

} // end of namespace splines
